from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

import numpy as np
import pandas as pd

from bridge.display.display_engine import DisplayEngine
from bridge.primitives.sample import Sample
from bridge.utils import optional_dependencies

if TYPE_CHECKING:
    from bridge.primitives.dataset import Dataset
    from bridge.primitives.element.element import Element
    from bridge.utils.data_objects import BoundingBox


class Panel(DisplayEngine):
    def __init__(self, bbox_format: str = "xyxy") -> None:
        assert bbox_format in ["xyxy", "xywh", "cxcywh"]
        self._bbox_format = bbox_format

    def show_element(self, element: Element, element_plot_kwargs: Dict[str, Any] | None = None):
        self._validate_dependencies()
        etype = element.etype
        if etype == "image":
            plot = self._plot_single_image(element)
        elif etype == "bbox":
            plot = self._plot_single_bbox(element)
        else:
            raise NotImplementedError(f"Invalid etype: {etype}")
        if element_plot_kwargs:
            plot = plot.opts(**element_plot_kwargs)
        return plot

    def show_sample(
        self,
        sample: Sample,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
    ):
        self._validate_dependencies()
        import holoviews as hv

        imgs = [self._plot_single_image(element) for element in sample.elements["image"]]
        if "bbox" in sample.elements:
            bboxes = self._plot_list_of_bbox_or_class_labels(sample.elements["bbox"])
        else:
            bboxes = hv.Overlay()
        if "class_label" in sample.elements:
            class_labels = self._plot_list_of_bbox_or_class_labels(sample.elements["class_label"])
        else:
            class_labels = hv.Overlay()
        for i in range(len(imgs)):
            imgs[i] = imgs[i] * bboxes * class_labels
        return hv.Layout(imgs)

    def show_dataset(
        self,
        dataset: Dataset,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
        dataset_plot_kwargs: Dict[str, Any] | None = None,
    ):
        self._validate_dependencies()
        import panel as pn

        sample_ids = dataset.sample_ids
        sample_ids_wig = pn.widgets.DiscreteSlider(name="Sample ID", options=sample_ids, value=sample_ids[0])

        @pn.depends(sample_ids_wig.param.value)
        def plot_sample_by_widget(sample_id):
            return self.show_sample(
                dataset.get(sample_id), element_plot_kwargs=element_plot_kwargs, sample_plot_kwargs=sample_plot_kwargs
            )

        return pn.Column(sample_ids_wig, plot_sample_by_widget)

    def _plot_single_image(self, element: Element):
        import holoviews as hv

        if element.category == "image":
            data: np.ndarray = element.data
        elif element.category == "torch":
            data: np.ndarray = element.data.permute(1, 2, 0).numpy()

        etype = element.etype
        h, w = data.shape[0], data.shape[1]
        if len(data.shape) == 2:
            # handling special case of grayscale images
            data = data[:, :, np.newaxis].repeat(3, axis=-1)
        img = hv.RGB(data[::-1, :, :], bounds=(0, 0, w, h)).opts(**self._default_kwargs(etype))
        return img

    def _plot_single_bbox(self, element: Element):
        import holoviews as hv

        etype = element.etype
        data: BoundingBox = element.data
        xyxy = self._extract_bbox_coords(data)
        plot = hv.Rectangles([[*xyxy, str(data.class_label)]], label=str(data.class_label)).opts(
            **self._default_kwargs(etype)
        )
        return plot

    def _plot_list_of_bbox_or_class_labels(self, elements: List[Element]):
        import holoviews as hv

        rectangle_list = []
        for element in elements:
            data: BoundingBox = element.data
            if element.etype == "bbox":
                xyxy = self._extract_bbox_coords(data)
                cl = data.class_label
            elif element.etype == "class_label":  # assume cls
                xyxy = [np.nan, np.nan, np.nan, np.nan]
                cl = data
            else:
                raise NotImplementedError()

            rectangle_list.append({"x0": xyxy[0], "y0": xyxy[1], "x1": xyxy[2], "y1": xyxy[3], "class": str(cl)})
        hv_df = pd.DataFrame.from_records(rectangle_list)
        if len(hv_df) == 0:
            return None
        else:
            plots = []
            for i, group in hv_df.groupby("class"):
                p = hv.Rectangles(group, label=i)
                plots.append(p)
            plots = hv.Overlay(plots).opts(hv.opts.Rectangles(**self._default_kwargs("bbox")))
            return plots

    def _extract_bbox_coords(self, data):
        if self._bbox_format == "xyxy":
            xyxy = data.coords
        elif self._bbox_format == "cxcywh":
            x, y, w, h = data.coords
            xyxy = np.clip(
                [x - w / 2, y - h / 2, x + w / 2, y + h / 2],
                a_min=0,
                a_max=None,
            )
        elif self._bbox_format == "xywh":
            x_min, y_min, w, h = data.coords
            xyxy = np.clip([x_min, y_min, x_min + w, y_min + h], a_min=0, a_max=None)
        else:
            raise NotImplementedError()
        return xyxy

    @staticmethod
    def _default_kwargs(etype: str) -> Dict[str, Any]:
        import holoviews as hv

        if etype == "image":
            return dict(aspect="equal", invert_yaxis=True, legend_position="left", xaxis=None, yaxis=None)
        elif etype == "bbox":
            return dict(fill_alpha=0.0, line_width=3, line_color=hv.Cycle("Category20"))

    @staticmethod
    def _validate_dependencies():
        with optional_dependencies("raise"):
            import holoviews  # noqa
            import panel  # noqa

        assert (
            holoviews.Store.current_backend == "bokeh"
        ), f"Holoviews backend: {holoviews.Store.current_backend} is not supported."



"""
pmap: Parallel Mapping Utility

This module provides a utility function, pmap, to apply a given function to a list of arguments in parallel.
It supports multiple backends for parallel processing, including 'concurrent', 'joblib', and 'dataloader' (PyTorch).

Usage:
    To use this utility, simply call the pmap function with the desired method, arguments list, and optional parameters.

Example:

    from pmap import pmap

    def square(x):
        return x * x

    results = pmap(square, [1, 2, 3, 4, 5])
    print(results)  # Output: [1, 4, 9, 16, 25]


Functions:
    - pmap: Parallel map function to apply a method to a list of arguments in parallel.
    - _pmap_dataloader: Internal function for parallel mapping using PyTorch DataLoader.
    - _pmap_concurrent_futures: Internal function for parallel mapping using concurrent.futures.
    - _pmap_joblib: Internal function for parallel mapping using Joblib.

"""

import concurrent.futures
import functools
import os
import pickle
from typing import Callable, Sequence

from tqdm import tqdm


def pmap(
    function: Callable,
    iterable: Sequence,
    progress_bar: bool = True,
    n_jobs: int = os.cpu_count(),
    backend: str = "concurrent",
) -> Sequence:
    """
    Apply a given method to a list of arguments in parallel, return outputs in the order of the input.

    Args:
        function (Callable): Method to apply.
        iterable (Sequence): Sequence of arguments.
        progress_bar (bool, optional): Show progress bar. Defaults to True.
        n_jobs (int, optional): Number of parallel jobs. Defaults to CPU count.
        backend (str, optional): Parallel backend ('concurrent', 'joblib', or 'dataloader' (PyTorch)).
        Defaults to 'concurrent'.

    Returns:
        Sequence: Sequence of results.
    """
    if n_jobs == 0:
        return [function(params) for params in (tqdm(iterable) if progress_bar else iterable)]
    else:
        if backend == "joblib":
            return _pmap_joblib(function, iterable, n_jobs, progress_bar)
        elif backend == "concurrent":
            return _pmap_concurrent_futures(function, iterable, n_jobs, progress_bar)
        elif backend == "dataloader":
            return _pmap_dataloader(function, iterable, n_jobs, progress_bar)


def _pmap_dataloader(function: Callable, iterable: Sequence, n_jobs: int, progress_bar: bool):
    """
    Internal function for parallel mapping using PyTorch DataLoader.

    Args:
        function (Callable): Method to apply.
        iterable (Sequence): Sequence of arguments.
        n_jobs (int): Number of parallel jobs.
        progress_bar (bool): Show progress bar.

    Returns:
        Sequence: Sequence of results.
    """
    from torch.utils.data import DataLoader, Dataset

    class DummyDataset(Dataset):
        def __init__(self, fn: Callable, args: Sequence):
            self._method = fn
            self._args_list = args

        def __getitem__(self, item):
            args = self._args_list[item]
            return self._method(args)

        def __len__(self):
            return len(self._args_list)

    def collate(data):
        return data

    ds = DummyDataset(function, iterable)
    dataloader = DataLoader(
        ds, num_workers=n_jobs, collate_fn=collate, batch_size=None, batch_sampler=None, prefetch_factor=10
    )
    if progress_bar:
        return list(tqdm(dataloader, total=len(ds), position=0, leave=True))
    else:
        return list(dataloader)


def _pmap_concurrent_futures(function: Callable, iterable: Sequence, n_jobs, progress_bar: bool):
    """
    Internal function for parallel mapping using concurrent.futures.

    Args:
        function (Callable): Method to apply.
        iterable (Sequence): Sequence of arguments.
        n_jobs (int): Number of parallel jobs.
        progress_bar (bool): Show progress bar.

    Returns:
        Sequence: Sequence of results.
    """
    try:
        pickle.dumps(function)
        helper_instance = function
    except pickle.PickleError:
        import dill

        function = dill.dumps(function)
        helper_instance = functools.partial(_helper, function)
    with concurrent.futures.ProcessPoolExecutor(max_workers=n_jobs) as p:
        if progress_bar:
            return p.map(helper_instance, tqdm(iterable, total=len(iterable)))
        else:
            return p.map(helper_instance, iterable)


def _helper(function, iterable, *args, **kwargs):
    """
    Helper function to unpickle method in parallel execution.

    Args:
        function (bytes): Pickled method.
        iterable: Input argument.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.

    Returns:
        Any: Result of applying the method.
    """
    import dill

    function = dill.loads(function)
    return function(iterable, *args, **kwargs)


def _pmap_joblib(function: Callable, iterable: Sequence, n_jobs: int, progress_bar: bool):
    """
    Internal function for parallel mapping using Joblib.

    Args:
        function (Callable): Method to apply.
        iterable (Sequence): Sequence of arguments.
        n_jobs (int): Number of parallel jobs.
        progress_bar (bool): Show progress bar.

    Returns:
        Sequence: Sequence of results.
    """
    from joblib import Parallel, delayed

    if progress_bar:
        return Parallel(n_jobs)(delayed(function)(args) for args in tqdm(iterable))
    else:
        return Parallel(n_jobs)(delayed(function)(args) for args in iterable)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

import pandas as pd
import panel as pn

from bridge.display import DisplayEngine
from bridge.primitives.dataset import SingularDataset
from bridge.primitives.sample.singular_sample import SingularSample

if TYPE_CHECKING:
    from bridge.primitives.element.element import Element


class Panel(DisplayEngine[SingularDataset, SingularSample]):
    def show_element(self, element: Element, element_plot_kwargs: Dict[str, Any] | None = None):
        if element.etype == "class_label":
            return pn.pane.Markdown(element.to_pd_series().to_frame().T.to_markdown())
        elif element.etype == "text":
            return pn.pane.Markdown(element.data)
        else:
            raise NotImplementedError()

    def show_sample(
        self,
        sample: SingularSample,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
    ):
        annotations_md = pd.DataFrame([ann.to_pd_series() for ann in sample.annotations["class_label"]]).to_markdown()
        text_display = pn.pane.Markdown(sample.data)
        return pn.Column("# Sample Text:", text_display, "# Annotations Data:", annotations_md)

    def show_dataset(
        self,
        dataset: SingularDataset,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
        dataset_plot_kwargs: Dict[str, Any] | None = None,
    ):
        sample_ids = dataset.sample_ids
        sample_ids_wig = pn.widgets.DiscreteSlider(name="Sample ID", options=sample_ids, value=sample_ids[0])

        @pn.depends(sample_ids_wig.param.value)
        def plot_sample_by_widget(sample_id):
            return self.show_sample(dataset.get(sample_id), element_plot_kwargs, sample_plot_kwargs)

        return pn.Column(sample_ids_wig, plot_sample_by_widget)



import abc
from pathlib import Path
from typing import Any

import numpy as np

from bridge.primitives.element.data.category_registry import register
from bridge.primitives.element.data.load_mechanism import LoadMechanism
from bridge.primitives.element.data.uri_components import URIComponents
from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE


class DataIO(abc.ABC):
    @property
    @abc.abstractmethod
    def category(self):
        pass

    @property
    @abc.abstractmethod
    def extension(self):
        pass

    @classmethod
    @abc.abstractmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        pass

    @classmethod
    @abc.abstractmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        pass


@register
class JPEGDataIO(DataIO):
    category = "image"
    extension = ".jpg"

    @classmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return np.array(url_or_data)  # assumes object is a PIL image or np.ndarray

        if url_or_data.scheme not in ["http", "https", "file", ""]:
            raise NotImplementedError("Only loading from local or http(s) URLs is supported for now.")
        from skimage.io import imread

        return imread(str(url_or_data))

    @classmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        data = np.array(data)
        if url is None:
            return LoadMechanism(data, cls.category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError("Only saving locally is supported for now.")
        from skimage.io import imsave

        path = Path(str(url)).expanduser()
        Path.mkdir(path.parent, parents=True, exist_ok=True)
        imsave(path, data)
        return LoadMechanism.from_url_string(str(path), cls.category)


@register
class TorchDataIO(DataIO):
    category = "torch"
    extension = ".pt"

    @classmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data  # assume that is already torch tensor
        import torch

        return torch.load(str(url_or_data))

    @classmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        if url is None:
            return LoadMechanism(data, cls.category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError("Only saving locally is supported for now.")
        import torch

        path = Path(str(url))

        Path.mkdir(path.parent, parents=True, exist_ok=True)
        torch.save(data, path)
        return LoadMechanism.from_url_string(str(url), cls.category)


@register
class NumpyDataIO(DataIO):
    category = "numpy"
    extension = ".npy"

    @classmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data
        return np.load(str(url_or_data))

    @classmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        raise NotImplementedError()


@register
class TextDataIO(DataIO):
    category = "text"
    extension = ".txt"

    @classmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data
        return open(str(url_or_data), "r").read()

    @classmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        raise NotImplementedError()


@register
class ObjDataIO(DataIO):
    category = "obj"
    extension = ".pkl"

    @classmethod
    def load(cls, url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data
        raise NotImplementedError()

    @classmethod
    def store(cls, data: Any, url: URIComponents | None) -> LoadMechanism:
        if url is None:
            return LoadMechanism(data, cls.category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError("Only saving locally is supported for now.")
        import pickle

        path = Path(str(url))

        Path.mkdir(path.parent, parents=True, exist_ok=True)
        with open(path, "wb") as f:
            pickle.dump(data, f)
        return LoadMechanism.from_url_string(str(url), cls.category)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, Hashable, List, Sequence

import pandas as pd
from typing_extensions import Self

from bridge.primitives.dataset.dataset import Dataset
from bridge.primitives.sample.singular_sample import SingularSample
from bridge.utils.constants import ELEMENT_COLS, INDICES, IS_SAMPLE_COL_NAME

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism
    from bridge.primitives.element.element import Element
    from bridge.primitives.sample.transform import SampleTransform


class SingularDataset(Dataset):
    """
    An annotated dataset is a popular use-case where a dataset is composed of samples (images, text, audio, video)
    and annotations (bboxes, captions, frames, labels)
    This implementation exposes `ds.elements` as two different views: `ds.samples` and `ds.annotations`, and the
    respective `select_<samples/annotations>`, `sort_<examples/annotations>`, `assign_<examples/annotations>` methods.
    """

    def __init__(
        self,
        samples: pd.DataFrame,
        annotations: pd.DataFrame,
        display_engine: DisplayEngine = None,
        cache_mechanisms: Dict[str, CacheMechanism | None] | None = None,
    ):
        assert (
            len(
                samples.index.get_level_values(ELEMENT_COLS.ID).intersection(
                    annotations.index.get_level_values(ELEMENT_COLS.ID)
                )
            )
            == 0
        ), "samples and annotations can't share ids"

        samples[IS_SAMPLE_COL_NAME] = True
        annotations[IS_SAMPLE_COL_NAME] = False
        elements = pd.concat([samples, annotations])
        super().__init__(elements, display_engine, cache_mechanisms)

    @property
    def samples(self) -> pd.DataFrame:
        return (
            self._elements.loc[self._elements[IS_SAMPLE_COL_NAME]]
            .dropna(axis="columns", how="all")
            .drop(columns=IS_SAMPLE_COL_NAME)
        )

    @property
    def annotations(self) -> pd.DataFrame:
        return (
            self._elements.loc[~self._elements[IS_SAMPLE_COL_NAME]]
            .dropna(axis="columns", how="all")
            .drop(columns=IS_SAMPLE_COL_NAME)
        )

    def iget(self, index: int) -> SingularSample:
        return SingularSample.from_sample(super().iget(index))

    def get(self, sample_id: Hashable) -> SingularSample:
        return SingularSample.from_sample(super().get(sample_id))

    def select_samples(self, selector: Callable[[pd.DataFrame, pd.DataFrame], Sequence]):
        selected = selector(self.samples, self.annotations)
        new_samples = self.samples.loc[selected]
        new_annotations = self.annotations.pipe(
            lambda df_: df_.loc[
                df_.index.get_level_values(ELEMENT_COLS.SAMPLE_ID).isin(
                    new_samples.index.get_level_values(ELEMENT_COLS.SAMPLE_ID)
                )
            ]
        )
        return SingularDataset(
            new_samples,
            new_annotations,
            display_engine=self._display_engine,
            cache_mechanisms=self._cache_mechanisms,
        )

    def select_annotations(self, selector: Callable[[pd.DataFrame, pd.DataFrame], Sequence]):
        selected = selector(self.samples, self.annotations)
        new_annotations = self.annotations.loc[selected]

        return SingularDataset(
            self.samples,
            new_annotations,
            display_engine=self._display_engine,
            cache_mechanisms=self._cache_mechanisms,
        )

    def assign_samples(self, **kwargs: Callable[[pd.DataFrame, pd.DataFrame], Sequence]) -> Self:
        values_dict = {name: assign_fn(self.samples, self.annotations) for name, assign_fn in kwargs.items()}
        new_samples = self.samples.assign(**values_dict)
        return SingularDataset(
            new_samples,
            self.annotations,
            display_engine=self._display_engine,
            cache_mechanisms=self._cache_mechanisms,
        )

    def assign_annotations(self, **kwargs: Callable[[pd.DataFrame, pd.DataFrame], Sequence]) -> Self:
        values_dict = {name: assign_fn(self.samples, self.annotations) for name, assign_fn in kwargs.items()}
        new_annotations = self.annotations.assign(**values_dict)
        return SingularDataset(
            self.samples,
            new_annotations,
            display_engine=self._display_engine,
            cache_mechanisms=self._cache_mechanisms,
        )

    def sort_samples(self, by: str, ascending: bool = True):
        new_samples = self.samples.sort_values(by=by, ascending=ascending)
        return SingularDataset(
            new_samples, self.annotations, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms
        )

    def sort_annotations(self, by: str, ascending: bool = True):
        new_annotations = self.annotations.sort_values(by=by, ascending=ascending)
        return SingularDataset(
            self.samples, new_annotations, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms
        )

    def transform_samples(
        self,
        transform: SampleTransform,
        map_fn=map,
        cache_mechanisms: Dict[str, CacheMechanism] | None = None,
        display_engine: DisplayEngine | None = None,
    ) -> Self:
        ds = super().transform_samples(
            transform, map_fn=map_fn, cache_mechanisms=cache_mechanisms, display_engine=display_engine
        )
        samples = (
            ds.elements.loc[ds.elements[IS_SAMPLE_COL_NAME]]
            .dropna(axis="columns", how="all")
            .drop(columns=IS_SAMPLE_COL_NAME)
        )
        annotations = (
            ds.elements.loc[~ds.elements[IS_SAMPLE_COL_NAME]]
            .dropna(axis="columns", how="all")
            .drop(columns=IS_SAMPLE_COL_NAME)
        )
        return SingularDataset(samples, annotations, display_engine=display_engine)

    @classmethod
    def from_lists(
        cls,
        samples_list: List[Element],
        annotations_list: List[Element],
        display_engine: DisplayEngine = None,
        cache_mechanisms: Dict[str, CacheMechanism | None] | None = None,
    ) -> Self:
        sample_records = [s.to_dict() for s in samples_list]
        annotation_records = [a.to_dict() for a in annotations_list]

        samples_df = pd.DataFrame(sample_records).set_index(INDICES)
        annotations_df = pd.DataFrame(annotation_records).set_index(INDICES)

        return cls(samples_df, annotations_df, display_engine=display_engine, cache_mechanisms=cache_mechanisms)



import warnings
from contextlib import contextmanager
from typing import List


@contextmanager
def optional_dependencies(error: str = "ignore", extras: List[str] | None = None):
    assert error in {"raise", "warn", "ignore"}
    try:
        yield None
    except ImportError as e:
        dep_name = e.name
        msg = _generate_msg(dep_name, extras)
        if error == "raise":
            raise ImportError(msg) from e
        if error == "warn":
            warnings.warn(msg, UserWarning)


def _generate_msg(dep_name: str, extras: List[str] | None):
    if extras is not None:
        extras_str = ", or alternatively install: " + ", ".join([f"bridge_ds[{extra}]" for extra in extras])
    else:
        extras_str = "."
    msg = f'Missing optional dependency "{dep_name}". Use pip or conda to install directly' + extras_str
    return msg



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List

from typing_extensions import Self

from bridge.primitives.sample import Sample
from bridge.utils.constants import IS_SAMPLE_COL_NAME

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism
    from bridge.primitives.element.element import Element
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE
    from bridge.primitives.sample.transform import SampleTransform


class SingularSample(Sample):
    def __init__(self, elements: List[Element] | Dict[str, List[Element]], display_engine: DisplayEngine = None):
        super().__init__(elements, display_engine)
        self._set_element_and_annotations()

    def _set_element_and_annotations(self):
        element = None
        found_element = False
        annotations = defaultdict(list)
        for etype, e_list in self.elements.items():
            for e in e_list:
                if e.metadata.get(IS_SAMPLE_COL_NAME) is True and found_element:
                    raise RuntimeError()
                if e.metadata.get(IS_SAMPLE_COL_NAME) is True:
                    found_element = True
                    element = e
                else:
                    annotations[etype].append(e)
        if not found_element:
            raise RuntimeError()

        self._element = element
        self._annotations = dict(annotations)

    @property
    def element(self) -> Element:
        return self._element

    @property
    def data(self) -> ELEMENT_DATA_TYPE:
        return self._element.data

    @property
    def annotations(self) -> Dict[str, List[Element]]:
        return self._annotations

    def transform(
        self,
        transform: SampleTransform,
        cache_mechanisms: Dict[str, CacheMechanism] | None = None,
        display_engine: DisplayEngine | None = None,
    ) -> Self:
        transformed_sample = super().transform(transform, cache_mechanisms, display_engine)
        return self.from_sample(transformed_sample)

    @classmethod
    def from_sample(cls, sample: Sample) -> Self:
        return cls(sample._elements, sample._display_engine)



from __future__ import annotations

from typing import Any, Dict, List


def validate_metadata(keys: List[str], metadata: Dict[str, Any] | None):
    if metadata is None:
        return
    dup_keys = set(metadata.keys()) & set(keys)
    assert dup_keys == set(), f"Metadata cannot contain reserved keys={dup_keys}"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Generic

from bridge.display.basic import SimplePrints
from bridge.primitives.types import D, S

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism


class DatasetProvider(ABC, Generic[D, S]):
    @abstractmethod
    def build_dataset(
        self,
        display_engine: DisplayEngine[D, S] = SimplePrints(),
        cache_mechanisms: Dict[str, CacheMechanism | None] | None = None,
    ) -> D:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Hashable

import pandas as pd

from bridge.primitives.element.data import category_registry
from bridge.primitives.element.data.uri_components import URIComponents

if TYPE_CHECKING:
    from bridge.primitives.element.data.load_mechanism import LoadMechanism
    from bridge.primitives.element.element import Element
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE


class CacheMechanism:
    def __init__(self, root_uri: URIComponents | None = None):
        self._elements = None
        self._root_uri = root_uri

    def set_elements_df(self, elements: pd.DataFrame):
        self._elements = elements

    def store(
        self,
        element: Element,
        data: ELEMENT_DATA_TYPE,
        as_category: str | None = None,
        should_update_elements: bool = False,
    ) -> LoadMechanism:
        if as_category is None:
            as_category = element.category
        assert category_registry.is_registered(as_category), f"Category {as_category} is not registered."
        uri = self._build_uri(element, as_category)
        new_provider = category_registry.store(data, uri, as_category)
        if should_update_elements and self._elements is not None:
            self._update_samples_with_new_provider(element.id, new_provider)
        return new_provider

    def _build_uri(self, element: Element, category: str) -> URIComponents | None:
        if self._root_uri is None:
            return None

        uri = URIComponents(
            scheme=self._root_uri.scheme,
            path=self._root_uri.path + f"/{element.id}{category_registry.extension(category)}",
        )
        return uri

    def _update_samples_with_new_provider(self, element_id: Hashable, new_provider: LoadMechanism):
        dic = new_provider.to_dict()
        self._elements.loc[(slice(None), element_id), list(dic.keys())] = dic.values()



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Dict

from bridge.display.text import Panel
from bridge.primitives.dataset.singular_dataset import SingularDataset
from bridge.primitives.element.data.load_mechanism import LoadMechanism
from bridge.primitives.element.element import Element
from bridge.primitives.sample.singular_sample import SingularSample
from bridge.providers.dataset_provider import DatasetProvider
from bridge.utils import download_and_extract_archive
from bridge.utils.data_objects import ClassLabel

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism


class LargeMovieReviewDataset(DatasetProvider[SingularDataset, SingularSample]):
    dataset_url = "https://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz"

    def __init__(self, root: str | os.PathLike, split: str = "train", download: bool = False):
        root = Path(root).expanduser()

        if download:
            if (root / "aclImdb_v1.tar.gz").exists():
                print("Archive file aclImdb_v1.tar.gz already exists, skipping download.")
            else:
                download_and_extract_archive(self.dataset_url, str(root))
        self._split_root = root / "aclImdb" / split

    def build_dataset(
        self,
        display_engine: DisplayEngine[SingularDataset, SingularSample] = Panel(),
        cache_mechanisms: Dict[str, CacheMechanism] = None,
    ) -> SingularDataset:
        samples = []
        annotations = []

        class_dir_list = [d for d in list(self._split_root.iterdir()) if d.is_dir()]
        for class_idx, class_dir in enumerate(sorted(class_dir_list)):
            for textfile in class_dir.iterdir():
                load_mechanism = LoadMechanism.from_url_string(str(textfile), "text")
                text_element = Element(
                    element_id=f"text_{textfile.stem}",
                    sample_id=textfile.stem,
                    etype="text",
                    load_mechanism=load_mechanism,
                )
                load_mechanism = LoadMechanism(ClassLabel(class_idx, class_dir.name), category="obj")
                label_element = Element(
                    element_id=f"label_{textfile.stem}",
                    sample_id=textfile.stem,
                    etype="class_label",
                    load_mechanism=load_mechanism,
                )
                samples.append(text_element)
                annotations.append(label_element)

        return SingularDataset.from_lists(
            samples, annotations, display_engine=display_engine, cache_mechanisms=cache_mechanisms
        )



import abc
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List


class StrEnum(str, Enum):
    def __str__(self):
        return self.value


class Dictable(ABC):
    @property
    @abstractmethod
    def keys(self) -> List[str]:
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, dic: Dict[str, Any], **kwargs):
        pass


class Displayable(abc.ABC):
    @abc.abstractmethod
    def show(self, **kwargs):
        pass



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Dict, List, Union

import numpy as np
import torch
from PIL import Image
from torchvision import tv_tensors
from torchvision.transforms import v2

from bridge.primitives.element.element import Element
from bridge.primitives.sample import Sample
from bridge.primitives.sample.transform.sample_transform import SampleTransform
from bridge.utils import optional_dependencies
from bridge.utils.data_objects import BoundingBox

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism


class TorchvisionV2Transform(SampleTransform):
    def __init__(
        self,
        transforms: Union[List[v2.Transform], v2.Compose],
        bbox_format="XYWH",
    ):
        if isinstance(transforms, v2.Compose):
            self._transforms = transforms
        else:
            self._transforms = v2.Compose(transforms)
        self._bbox_format = bbox_format

    def __call__(
        self, sample: Sample, cache_mechanisms: Dict[str, CacheMechanism], display_engine: DisplayEngine | None
    ) -> Sample:
        # Extract and validate required elements
        image_element, bbox_elements = self._extract_elements(sample)

        # Convert to torchvision v2 format
        image_tensor, bbox_tensor, labels_tensor = self._convert_to_tv_tensors(image_element, bbox_elements)

        # Apply transforms
        transformed_image, transformed_bboxes, transformed_labels = self._apply_transforms(
            image_tensor, bbox_tensor, labels_tensor
        )

        # Reconstruct elements with transformed data
        new_elements = self._reconstruct_elements(
            sample.elements,
            image_element,
            bbox_elements,
            transformed_image,
            transformed_bboxes,
            transformed_labels,
            cache_mechanisms,
        )

        return Sample(elements=new_elements, display_engine=display_engine)

    def _extract_elements(self, sample: Sample) -> tuple[Element, List[Element]]:
        """Extract and validate image and bbox elements from sample."""
        if "image" not in sample.elements:
            raise ValueError("TorchvisionV2Transform requires 'image' element")

        image_element = sample.elements["image"][0]  # Single image element
        bbox_elements = sample.elements.get("bbox", [])  # List of bbox elements (optional)

        return image_element, bbox_elements

    def _convert_to_tv_tensors(
        self, image_element: Element, bbox_elements: List[Element]
    ) -> tuple[Any, tv_tensors.BoundingBoxes, torch.Tensor]:
        """Convert elements to torchvision v2 tensor format."""
        # Convert image to tv_tensors format
        image_data = image_element.data
        if image_element.category == "image":
            # convert to PIL image
            image_data = Image.fromarray(image_data)
            W, H = image_data.size
        elif image_element.category == "torch":
            image_data = tv_tensors.Image(image_data)
            W, H = image_data.shape[1], image_data.shape[2]

        # Convert bboxes to tv_tensors format with artificial labels
        bbox_coords = []
        bbox_labels = []
        for i, bbox_element in enumerate(bbox_elements):
            bbox_data: BoundingBox = bbox_element.data
            bbox_coords.append(bbox_data.coords)
            bbox_labels.append(i)  # Artificial label as index

        if bbox_coords:
            bbox_tensor = tv_tensors.BoundingBoxes(
                torch.stack([torch.from_numpy(coord) for coord in bbox_coords]),
                format=self._bbox_format,
                canvas_size=(H, W),
            )
            labels_tensor = torch.tensor(bbox_labels)
        else:
            bbox_tensor = tv_tensors.BoundingBoxes(
                torch.empty((0, 4)),
                format=self._bbox_format,
                canvas_size=(H, W),
            )
            labels_tensor = torch.empty(0, dtype=torch.long)

        return image_data, bbox_tensor, labels_tensor

    def _apply_transforms(
        self, image_data: Any, bbox_tensor: tv_tensors.BoundingBoxes, labels_tensor: torch.Tensor
    ) -> tuple[Any, Any, torch.Tensor]:
        """Apply torchvision v2 transforms to tensors."""
        # Apply transforms to image, bboxes, and labels together
        output_dict = self._transforms({"image": image_data, "boxes": bbox_tensor, "labels": labels_tensor})
        transformed_image = output_dict["image"]
        transformed_bboxes = output_dict["boxes"]
        transformed_labels = output_dict["labels"]
        transformed_labels = labels_tensor

        return transformed_image, transformed_bboxes, transformed_labels

    def _reconstruct_elements(
        self,
        original_elements: Dict[str, List[Element]],
        image_element: Element,
        bbox_elements: List[Element],
        transformed_image: Any,
        transformed_bboxes: Any,
        transformed_labels: torch.Tensor,
        cache_mechanisms: Dict[str, CacheMechanism],
    ) -> Dict[str, List[Element]]:
        """Reconstruct elements dict with transformed data."""
        # Start with a copy of all original elements to preserve untransformed ones
        new_elements = copy.deepcopy(original_elements)

        # Update the image element with transformed data
        new_image_element = self._create_transformed_element(image_element, transformed_image, cache_mechanisms)
        new_elements["image"] = [new_image_element]

        # Update bbox elements for surviving bboxes using the transformed labels
        # (SanitizeBoundingBoxes should be included in the user's transform pipeline)
        # The labels contain the original indices of the surviving bboxes
        surviving_bbox_elements = []
        for i, bbox_coords in enumerate(transformed_bboxes):
            original_idx = transformed_labels[i].item()  # Get the original index from the label
            if original_idx < len(bbox_elements):
                original_bbox_element = bbox_elements[original_idx]
                bbox_coords_np = bbox_coords.numpy()
                new_bbox_data = BoundingBox(coords=bbox_coords_np, class_label=original_bbox_element.data.class_label)

                new_bbox_element = self._create_transformed_element(
                    original_bbox_element, new_bbox_data, cache_mechanisms
                )
                surviving_bbox_elements.append(new_bbox_element)

        # Replace the bbox elements with only the surviving ones
        new_elements["bbox"] = surviving_bbox_elements

        return new_elements

    def _create_transformed_element(
        self, original_element: Element, transformed_data: Any, cache_mechanisms: Dict[str, CacheMechanism]
    ) -> Element:
        """Create a new element with transformed data"""
        if original_element.etype == "image":
            if isinstance(transformed_data, np.ndarray) or isinstance(transformed_data, Image.Image):
                new_category = "image"
            else:
                with optional_dependencies(error="raise"):
                    import torch
                if isinstance(transformed_data, torch.Tensor):
                    new_category = "torch"
                else:
                    raise NotImplementedError(f"Invalid data type: {type(transformed_data)}")
        elif original_element.etype == "bbox":
            new_category = "obj"
        else:
            raise NotImplementedError(f"Unsupported element type: {original_element.etype}")

        provider = cache_mechanisms[original_element.etype].store(
            original_element, transformed_data, as_category=new_category, should_update_elements=False
        )

        return Element(
            element_id=original_element.id,
            etype=original_element.etype,
            load_mechanism=provider,
            sample_id=original_element.sample_id,
            metadata=original_element.metadata,
        )



from __future__ import annotations

from typing import TypeVar

from bridge.primitives.dataset import Dataset
from bridge.primitives.sample import Sample

D = TypeVar("D", bound=Dataset)
S = TypeVar("S", bound=Sample)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from bridge.primitives.element.data.uri_components import URIComponents
from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE

if TYPE_CHECKING:
    from bridge.primitives.element.data.load_mechanism import LoadMechanism

REGISTRY = {}


def is_registered(category: str) -> bool:
    return category in REGISTRY


def list_registered_categories() -> list[str]:
    return list(REGISTRY.keys())


def register(cls):
    if cls.category in REGISTRY:
        raise ValueError(f"Category {cls.category} is already registered.")
    REGISTRY[cls.category] = cls
    return cls


def store(data: Any, url: URIComponents | None, category: str) -> LoadMechanism:
    return REGISTRY[category].store(data, url)


def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE, category: str) -> ELEMENT_DATA_TYPE:
    return REGISTRY[category].load(url_or_data)


def extension(category: str) -> str:
    return REGISTRY[category].extension


# register default data io classes
import bridge.primitives.element.data.data_io  # noqa



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from typing_extensions import Self

from bridge.primitives.element.data import category_registry
from bridge.primitives.element.data.uri_components import URIComponents
from bridge.utils import Dictable
from bridge.utils.constants import ELEMENT_COLS

if TYPE_CHECKING:
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE


class LoadMechanism(Dictable):
    keys = ELEMENT_COLS.LOAD_MECHANISM.list()

    def __init__(self, url_or_data: URIComponents | ELEMENT_DATA_TYPE, category: str) -> None:
        assert category_registry.is_registered(category), f"Category {category} is not registered."
        self._url_or_data = url_or_data
        self._category = category

    @property
    def url_or_data(self) -> URIComponents | ELEMENT_DATA_TYPE:
        return self._url_or_data

    @property
    def category(self) -> str:
        return self._category

    def load_data(self) -> Any:
        return category_registry.load(self._url_or_data, self._category)

    def to_dict(self) -> Dict[str, Any]:
        return {
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: self.url_or_data,
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: self.category,
        }

    @classmethod
    def from_dict(cls, dic: Dict[str, Any], **kwargs) -> Self:
        assert set(dic.keys()).issuperset(set(cls.keys))
        return cls(
            url_or_data=dic[ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA],
            category=dic[ELEMENT_COLS.LOAD_MECHANISM.CATEGORY],
        )

    @classmethod
    def from_url_string(cls, url_string: str, category: str) -> Self:
        components = URIComponents.from_str(url_string)
        return cls(components, category)



from dataclasses import dataclass

import numpy as np


@dataclass
class ClassLabel:
    class_idx: int
    class_name: str | None = None

    def __post_init__(self):
        if self.class_name is None:
            self.class_name = str(self.class_idx)

    def __str__(self):
        return self.class_name

    def __hash__(self):
        return hash(f"{self.class_idx},{self.class_name}")

    def __eq__(self, other):
        if isinstance(other, str):
            return self.class_name == other
        else:
            return hash(self) == hash(other)


@dataclass
class BoundingBox:
    coords: np.ndarray
    class_label: ClassLabel | None = None

    def __post_init__(self):
        self.coords = self.coords.squeeze()
        assert self.coords.shape == (4,), f"Input coords to BoundingBox is wrong, {self.coords.shape}. Expected: 4"

    def __str__(self):
        if self.class_label is not None:
            return f"BoundingBox(class_name={self.class_label},coords={self.coords}"
        else:
            return f"BoundingBox(coords={self.coords})"


@dataclass
class Keypoint:
    coords: np.ndarray
    class_label: int | None = None
    class_name: str | None = None

    def __post_init__(self):
        self.coords = self.coords.squeeze()
        assert self.coords.shape == (2,), f"Input coords to Keypoint is wrong, {self.coords.shape}. Expected: 2"

    def __str__(self):
        if self.class_name is not None:
            return f"Keypoint(class_name={self.class_name},coords={self.coords}"
        else:
            return f"Keypoint(coords={self.coords})"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache_mechanism import CacheMechanism
    from bridge.primitives.sample import Sample


class SampleTransform(ABC):
    @abstractmethod
    def __call__(
        self,
        sample: Sample,
        cache_mechanisms: Dict[str, CacheMechanism] | None,
        display_engine: DisplayEngine | None,
    ) -> Sample:
        pass



import abc
from typing import Callable, Dict, Sequence

import pandas as pd
from typing_extensions import Self


class TableAPI(abc.ABC):
    @property
    @abc.abstractmethod
    def elements(self) -> pd.DataFrame:
        pass

    @abc.abstractmethod
    def select(self, selector: Callable) -> Self:
        pass

    @abc.abstractmethod
    def assign(self, **kwargs: Dict[str, Callable[[pd.DataFrame], Sequence]]) -> Self:
        pass

    @abc.abstractmethod
    def sort(self, by: str, ascending: bool = True) -> Self:
        pass
